@@ -27,17 +27,21 @@ use std::time::Instant;
 use tokio_native_tls::TlsConnector;
 
 #[derive(Builder, Deserialize, Clone, Debug)]
-#[builder(build_fn(validate = "Self::validate"))]
-pub struct HttpBenchAdapter {
-    url: Vec<String>,
+pub struct HttpClientConfig {
     #[builder(default)]
     ignore_cert: bool,
     #[builder(default)]
     conn_reuse: bool,
     #[builder(default)]
-    verbose: bool,
-    #[builder(default)]
     http2_only: bool,
+    #[builder(default)]
+    pub stop_on_errors: Vec<u16>,
+}
+
+#[derive(Builder, Deserialize, Clone, Debug)]
+#[builder(build_fn(validate = "Self::validate"))]
+pub struct HttpRequest {
+    url: Vec<String>,
     #[builder(default = "\"GET\".to_string()")]
     method: String,
     #[builder(default)]
@@ -46,6 +50,12 @@ pub struct HttpBenchAdapter {
     body: Vec<u8>,
 }
 
+#[derive(Builder, Deserialize, Clone, Debug)]
+pub struct HttpBenchAdapter {
+    config: HttpClientConfig,
+    request: HttpRequest,
+}
+
 #[cfg(feature = "tls")]
 type ProtocolConnector = HttpsConnector<HttpConnector>;
 #[cfg(not(feature = "tls"))]
@@ -73,7 +83,7 @@ impl HttpBenchAdapter {
     #[cfg(feature = "tls-native")]
     fn build_tls_connector(&self) -> TlsConnector {
         let mut native_tls_builder = native_tls::TlsConnector::builder();
-        if self.ignore_cert {
+        if self.config.ignore_cert {
             native_tls_builder.danger_accept_invalid_certs(true);
         }
 
@@ -99,30 +109,29 @@ impl BenchmarkProtocolAdapter for HttpBenchAdapter {
     type Client = hyper::Client<ProtocolConnector>;
 
     fn build_client(&self) -> Result<Self::Client, String> {
-        let mut client_builder = hyper::Client::builder();
-
-        if self.http2_only {
-            client_builder.http2_only(true);
-        }
-
-        if !self.conn_reuse {
-            client_builder.pool_max_idle_per_host(0);
-        }
-
-        Ok(client_builder.build(self.build_connector()))
+        Ok(hyper::Client::builder()
+            .http2_only(self.config.http2_only)
+            .pool_max_idle_per_host(if !self.config.conn_reuse {
+                0
+            } else {
+                usize::MAX
+            })
+            .build(self.build_connector()))
     }
 
     async fn send_request(&self, client: &Self::Client) -> RequestStats {
         let start = Instant::now();
-
-        let request = self.build_request();
-
+        let request = self.request.build_request();
         let response = client.request(request).await;
 
         match response {
             Ok(r) => {
                 let status = r.status().to_string();
                 let success = r.status().is_success();
+
+                let fatal_error =
+                    !success && self.config.stop_on_errors.contains(&r.status().as_u16());
+
                 let mut stream = r.into_body();
                 let mut total_size = 0;
                 while let Some(item) = stream.next().await {
@@ -137,6 +146,7 @@ impl BenchmarkProtocolAdapter for HttpBenchAdapter {
                     .status(status)
                     .is_success(success)
                     .duration(Instant::now().duration_since(start))
+                    .fatal_error(fatal_error)
                     .build()
                     .expect("RequestStatsBuilder failed")
             }
@@ -148,6 +158,7 @@ impl BenchmarkProtocolAdapter for HttpBenchAdapter {
                     .status(status)
                     .is_success(false)
                     .duration(Instant::now().duration_since(start))
+                    .fatal_error(false)
                     .build()
                     .expect("RequestStatsBuilder failed")
             }
@@ -155,14 +166,13 @@ impl BenchmarkProtocolAdapter for HttpBenchAdapter {
     }
 }
 
-impl HttpBenchAdapter {
+impl HttpRequest {
     fn build_request(&self) -> Request<Body> {
         let method =
             Method::from_str(&self.method.clone()).expect("Method must be valid at this point");
 
-        let mut request_builder = Request::builder()
-            .method(method)
-            .uri(&self.url[thread_rng().gen_range(0..self.url.len())].clone());
+        let uri = &self.url[thread_rng().gen_range(0..self.url.len())];
+        let mut request_builder = Request::builder().method(method).uri(uri.clone());
 
         if !self.headers.is_empty() {
             for (key, value) in self.headers.iter() {
@@ -180,12 +190,18 @@ impl HttpBenchAdapter {
         } else {
             request_builder
                 .body(Body::empty())
+                .map_err(|e| {
+                    println!(
+                        "Cannot create url {}, headers: {:?}. Error: {}",
+                        uri, self.headers, e
+                    );
+                })
                 .expect("Error building Request")
         }
     }
 }
 
-impl HttpBenchAdapterBuilder {
+impl HttpRequestBuilder {
     /// Validate request is going to be built from the given settings
     fn validate(&self) -> Result<(), String> {
         if let Some(ref m) = self.method {
@@ -199,7 +215,9 @@ impl HttpBenchAdapterBuilder {
 #[cfg(test)]
 mod tests {
     use crate::bench_run::BenchmarkProtocolAdapter;
-    use crate::http_bench_session::{HttpBenchAdapter, HttpBenchAdapterBuilder};
+    use crate::http_bench_session::{
+        HttpBenchAdapter, HttpBenchAdapterBuilder, HttpClientConfigBuilder, HttpRequestBuilder,
+    };
     use mockito::mock;
     use mockito::Matcher::Exact;
     use std::time::Duration;
@@ -220,11 +238,17 @@ mod tests {
         let url = mockito::server_url().to_string();
         println!("Url: {}", url);
         let http_bench = HttpBenchAdapterBuilder::default()
-            .url(vec![format!("{}/1", url)])
-            .headers(vec![
-                ("x-header".to_string(), "value1".to_string()),
-                ("x-another-header".to_string(), "value2".to_string()),
-            ])
+            .request(
+                HttpRequestBuilder::default()
+                    .url(vec![format!("{}/1", url)])
+                    .headers(vec![
+                        ("x-header".to_string(), "value1".to_string()),
+                        ("x-another-header".to_string(), "value2".to_string()),
+                    ])
+                    .build()
+                    .unwrap(),
+            )
+            .config(HttpClientConfigBuilder::default().build().unwrap())
             .build()
             .unwrap();
 
@@ -252,13 +276,19 @@ mod tests {
         let url = mockito::server_url().to_string();
         println!("Url: {}", url);
         let http_bench = HttpBenchAdapterBuilder::default()
-            .url(vec![format!("{}/1", url)])
-            .method("PUT".to_string())
-            .headers(vec![
-                ("x-header".to_string(), "value1".to_string()),
-                ("x-another-header".to_string(), "value2".to_string()),
-            ])
-            .body("abcd".as_bytes().to_vec())
+            .request(
+                HttpRequestBuilder::default()
+                    .url(vec![format!("{}/1", url)])
+                    .method("PUT".to_string())
+                    .headers(vec![
+                        ("x-header".to_string(), "value1".to_string()),
+                        ("x-another-header".to_string(), "value2".to_string()),
+                    ])
+                    .body("abcd".as_bytes().to_vec())
+                    .build()
+                    .unwrap(),
+            )
+            .config(HttpClientConfigBuilder::default().build().unwrap())
             .build()
             .unwrap();
 
@@ -286,13 +316,19 @@ mod tests {
         let url = mockito::server_url().to_string();
         println!("Url: {}", url);
         let http_bench = HttpBenchAdapterBuilder::default()
-            .url(vec![format!("{}/1", url)])
-            .method("POST".to_string())
-            .headers(vec![
-                ("x-header".to_string(), "value1".to_string()),
-                ("x-another-header".to_string(), "value2".to_string()),
-            ])
-            .body("abcd".as_bytes().to_vec())
+            .request(
+                HttpRequestBuilder::default()
+                    .url(vec![format!("{}/1", url)])
+                    .method("POST".to_string())
+                    .headers(vec![
+                        ("x-header".to_string(), "value1".to_string()),
+                        ("x-another-header".to_string(), "value2".to_string()),
+                    ])
+                    .body("abcd".as_bytes().to_vec())
+                    .build()
+                    .unwrap(),
+            )
+            .config(HttpClientConfigBuilder::default().build().unwrap())
             .build()
             .unwrap();
 
@@ -317,7 +353,13 @@ mod tests {
         let url = mockito::server_url().to_string();
         println!("Url: {}", url);
         let http_bench = HttpBenchAdapterBuilder::default()
-            .url(vec![format!("{}/1", url)])
+            .request(
+                HttpRequestBuilder::default()
+                    .url(vec![format!("{}/1", url)])
+                    .build()
+                    .unwrap(),
+            )
+            .config(HttpClientConfigBuilder::default().build().unwrap())
             .build()
             .unwrap();
 
@@ -342,8 +384,18 @@ mod tests {
         let url = mockito::server_url().to_string();
         println!("Url: {}", url);
         let http_bench: HttpBenchAdapter = HttpBenchAdapterBuilder::default()
-            .url(vec![format!("{}/1", url)])
-            .http2_only(true)
+            .request(
+                HttpRequestBuilder::default()
+                    .url(vec![format!("{}/1", url)])
+                    .build()
+                    .unwrap(),
+            )
+            .config(
+                HttpClientConfigBuilder::default()
+                    .http2_only(true)
+                    .build()
+                    .unwrap(),
+            )
             .build()
             .unwrap();
 
